# -*- coding: utf-8 -*-

"""
"""

import asyncio
import ctypes
import hashlib
import logging
import os
import subprocess
import sys

logger = logging.getLogger("wsterm")


class ConnectWebsocketServerFailed(RuntimeError):
    pass


class ConnectionClosedError(RuntimeError):
    pass


class AsyncFileDescriptor(object):
    """Async File Descriptor"""

    def __init__(self, fd):
        self._loop = asyncio.get_event_loop()
        self._fd = fd
        self._event = asyncio.Event()
        self._buffer = b""
        self._loop.add_reader(self._fd, self.read_callback)
        self._closed = False

    def close(self):
        self._loop.remove_reader(self._fd)

    async def read(self, bytes=4096):
        await self._event.wait()
        self._event.clear()
        buffer = self._buffer
        self._buffer = b""
        return buffer

    def write(self, buffer):
        os.write(self._fd, buffer)

    def read_callback(self, *args):
        try:
            buffer = os.read(self._fd, 4096)
        except OSError:
            self.close()
            self._closed = True
            self._event.set()
            return

        self._buffer += buffer
        self._event.set()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, exc_trackback):
        self.close()
        self._closed = True


class Process(object):
    def __init__(self, pid):
        self._pid = pid
        self._returncode = None
        safe_ensure_future(self._wait_for_exit())

    @property
    def returncode(self):
        return self._returncode

    async def _wait_for_exit(self):
        while True:
            try:
                pid, returncode = os.waitpid(self._pid, os.WNOHANG)
            except ChildProcessError:
                logger.warn(
                    "[%s] Process %d already exited" % (self.__class__.__name__, pid)
                )
                self._returncode = -1
                break

            if not pid:
                await asyncio.sleep(0.01)
            else:
                self._returncode = returncode
                break


class UnixStdIn(object):
    def __init__(self, stdin=None):
        self._fd = stdin or sys.stdin.fileno()
        import termios

        self._settings = termios.tcgetattr(self._fd)

    def __enter__(self):
        import tty

        tty.setraw(self._fd)
        return self

    def __exit__(self, exc_type, exc_value, exc_trackback):
        import termios

        termios.tcsetattr(self._fd, termios.TCSADRAIN, self._settings)

    def fileno(self):
        return self._fd

    def read(self, n):
        return os.read(self._fd, n)


def safe_ensure_future(coro, loop=None):
    loop = loop or asyncio.get_event_loop()
    fut = loop.create_future()

    async def _wrap():
        try:
            fut.set_result(await coro)
        except Exception as e:
            fut.set_exception(e)

    asyncio.ensure_future(_wrap())
    return fut


def enable_native_ansi():
    """Enables native ANSI sequences in console. Windows 10 only.
    Returns whether successful.
    """
    import ctypes.wintypes

    ENABLE_VIRTUAL_TERMINAL_PROCESSING = 0x04

    out_handle = ctypes.windll.kernel32.GetStdHandle(subprocess.STD_OUTPUT_HANDLE)

    # GetConsoleMode fails if the terminal isn't native.
    mode = ctypes.wintypes.DWORD()
    if ctypes.windll.kernel32.GetConsoleMode(out_handle, ctypes.byref(mode)) == 0:
        return False

    if not (mode.value & ENABLE_VIRTUAL_TERMINAL_PROCESSING):
        if (
            ctypes.windll.kernel32.SetConsoleMode(
                out_handle, mode.value | ENABLE_VIRTUAL_TERMINAL_PROCESSING
            )
            == 0
        ):
            print(
                "kernel32.SetConsoleMode to enable ANSI sequences failed",
                file=sys.stderr,
            )
            return False

    return True


def make_short_hash(string):
    if not isinstance(string, bytes):
        string = string.encode("utf8")
    return hashlib.sha1(string).hexdigest()[:8]


def diff(data1, data2):
    """Get diff of data2 and data1"""
    result = {}
    for key in data1:
        if key not in data2:
            result[key] = data1[key]
        else:
            res = None
            if isinstance(data1[key], dict) and isinstance(data2[key], dict):
                res = diff(data1[key], data2[key])
            elif data1[key] != data2[key]:
                res = data1[key]
            if res:
                result[key] = res
    for key in data2:
        if key not in data1:
            # 已删除的节点
            result[key] = "-"
    return result


def win32_daemon():
    cmdline = []
    for it in sys.argv:
        if it not in ("-d", "--daemon"):
            cmdline.append(it)

    DETACHED_PROCESS = 8
    subprocess.Popen(cmdline, creationflags=DETACHED_PROCESS, close_fds=True)



# -*- coding: utf-8 -*-

import asyncio
import os
import sys

import tornado.web
import tornado.websocket

from . import proto, shell, utils, workspace


class WebSocketProtocol(tornado.websocket.WebSocketProtocol13):
    async def accept_connection(self, handler):
        if self.handler.check_permission():
            await super(WebSocketProtocol, self).accept_connection(handler)
        else:
            handler.set_status(403)
            log_msg = "Authorization Failed"
            handler.finish(log_msg)


class WSTerminalServerHandler(tornado.websocket.WebSocketHandler):
    """Websocket Terminal Server Handler"""

    token = None

    def __init__(self, *args, **kwargs):
        super(WSTerminalServerHandler, self).__init__(*args, **kwargs)
        self._buffer = b""
        self._workspace = None
        self._shell = None
        self._sequence = 0x10000

    def check_permission(self):
        if self.token:
            auth = self.request.headers.get("Authorization", "")
            if auth.startswith("Token "):
                return auth.split()[-1].strip() == self.token
            else:
                return False
        return True

    def get_websocket_protocol(self):
        """Override to connect target server"""
        websocket_version = self.request.headers.get("Sec-WebSocket-Version")
        if websocket_version in ("7", "8", "13"):
            params = tornado.websocket._WebSocketParams(
                ping_interval=self.ping_interval,
                ping_timeout=self.ping_timeout,
                max_message_size=self.max_message_size,
                compression_options=self.get_compression_options(),
            )
            return WebSocketProtocol(self, mask_outgoing=True, params=params)

    async def on_message(self, message):
        self._buffer += message
        packet, self._buffer = proto.TransportPacket.deserialize(self._buffer)
        if packet:
            await self.handle_request(packet.message)

    async def send_request(self, command, **kwargs):
        self._sequence += 1
        data = {
            "command": command,
            "type": proto.EnumPacketType.REQUEST,
            "id": self._sequence,
        }
        data.update(kwargs)
        packet = proto.TransportPacket(data)
        return await self.write_message(packet.serialize(), True)

    async def send_response(self, request, code=0, message=None, **kwargs):
        data = {
            "command": request["command"],
            "type": proto.EnumPacketType.RESPONSE,
            "id": request["id"],
            "code": code,
            "message": message or "",
        }
        data.update(kwargs)
        packet = proto.TransportPacket(data)
        return await self.write_message(packet.serialize(), True)

    async def handle_request(self, request):
        utils.logger.debug(
            "[%s][Request][%d][%s] %s"
            % (
                self.__class__.__name__,
                request.get("id", 0),
                request.get("command"),
                str(request)[:200],
            )
        )
        if request["command"] == proto.EnumCommand.SYNC_WORKSPACE:
            worksapce_id = request["workspace"]
            workspace_path = os.path.join(
                os.environ.get("WSTERM_WORKSPACE", os.environ.get("TEMP", "/tmp")),
                worksapce_id,
            )
            self._workspace = workspace.Workspace(workspace_path)
            data = self._workspace.snapshot()
            await self.send_response(
                request, data=data,
            )
        elif self._workspace and request["command"] == proto.EnumCommand.WRITE_FILE:
            utils.logger.info(
                "[%s] Update file %s" % (self.__class__.__name__, request["path"])
            )
            self._workspace.write_file(
                request["path"], request["data"], request["overwrite"]
            )
        elif self._workspace and request["command"] == proto.EnumCommand.REMOVE_FILE:
            utils.logger.info(
                "[%s] Remove file %s" % (self.__class__.__name__, request["path"])
            )
            self._workspace.remove_file(request["path"])
        elif self._workspace and request["command"] == proto.EnumCommand.CREATE_DIR:
            utils.logger.info(
                "[%s] Create directory %s" % (self.__class__.__name__, request["path"])
            )
            self._workspace.create_directory(request["path"])
        elif self._workspace and request["command"] == proto.EnumCommand.REMOVE_DIR:
            utils.logger.info(
                "[%s] Remove directory %s" % (self.__class__.__name__, request["path"])
            )
            self._workspace.remove_directory(request["path"])
        elif self._workspace and request["command"] == proto.EnumCommand.MOVE_ITEM:
            utils.logger.info(
                "[%s] Move item %s to %s"
                % (self.__class__.__name__, request["src_path"], request["dst_path"])
            )
            self._workspace.move_item(request["src_path"], request["dst_path"])
        elif request["command"] == proto.EnumCommand.CREATE_SHELL:
            utils.logger.info(
                "[%s] Create shell (%d, %d)"
                % (self.__class__.__name__, *request["size"])
            )
            if self._shell:
                await self.send_response(request, code=-1, message="Shell is created")
            else:
                await self.send_response(request, platform=sys.platform)
                shell_workspace = os.getcwd()
                if self._workspace:
                    shell_workspace = self._workspace.path
                asyncio.ensure_future(
                    self.spawn_shell(shell_workspace, request["size"])
                )
        elif request["command"] == proto.EnumCommand.WRITE_STDIN:
            if not self._shell:
                await self.send_response(request, code=-1, message="Shell not create")
            else:
                utils.logger.debug(
                    "[%s] Input %s" % (self.__class__.__name__, request["buffer"])
                )
                self._shell.write(request["buffer"])

    async def write_shell_stdout(self, buffer):
        utils.logger.debug("[%s] Output %s" % (self.__class__.__name__, buffer))
        await self.send_request(proto.EnumCommand.WRITE_STDOUT, buffer=buffer)

    async def spawn_shell(self, workspace, size):
        utils.logger.info(
            "[%s] Spawn new shell (%d, %d)"
            % (self.__class__.__name__, size[0], size[1])
        )
        self._shell = shell.Shell(workspace, size)
        proc, _, _, _ = await self._shell.create()
        tasks = [None]
        if self._shell.stderr:
            tasks.append(None)
        while proc.returncode is None:
            if tasks[0] is None:
                tasks[0] = utils.safe_ensure_future(self._shell.stdout.read(4096))
            if self._shell.stderr and tasks[1] is None:
                tasks[1] = utils.safe_ensure_future(self._shell.stderr.read(4096))

            done_tasks, _ = await asyncio.wait(
                tasks, return_when=asyncio.FIRST_COMPLETED
            )

            for task in done_tasks:
                index = tasks.index(task)
                assert index >= 0
                tasks[index] = None
                buffer = task.result()
                if not buffer:
                    await asyncio.sleep(0.01)
                    break

                await self.write_shell_stdout(buffer)
        utils.logger.warn("[%s] Shell process exit" % self.__class__.__name__)
        await self.send_request(proto.EnumCommand.SHELL_EXIT)
        self._shell = None

    def on_connection_close(self):
        pass


def start_server(listen_address, path, token=None):
    utils.logger.info("Websocket server listening at %s:%d" % listen_address)
    WSTerminalServerHandler.token = token
    handlers = [
        (path, WSTerminalServerHandler),
    ]
    app = tornado.web.Application(handlers)
    app.listen(listen_address[1], listen_address[0])



# -*- coding: utf-8 -*-

"""Websocket Terminal Client
"""

import asyncio
import os
import socket
import sys
import time

import tornado.httputil
import tornado.websocket

from . import proto, utils, workspace


class WSTerminalConnection(tornado.websocket.WebSocketClientConnection):
    def __init__(self, url, headers=None, timeout=15, handler=None):
        self._url = url
        self.__timeout = timeout
        self._handler = handler
        self._connected = None
        self._closed = False
        compression_options = None
        if isinstance(headers, dict):
            headers = tornado.httputil.HTTPHeaders(headers)
        request = tornado.httpclient.HTTPRequest(
            self._url, headers=headers, connect_timeout=timeout, request_timeout=timeout
        )
        request = tornado.httpclient._RequestProxy(
            request, tornado.httpclient.HTTPRequest._DEFAULTS
        )
        super(WSTerminalConnection, self).__init__(
            request,
            on_message_callback=self.on_message,
            compression_options=compression_options,
        )
        self._buffer = b""
        self._queue = asyncio.Queue()
        self._rsp_map = {}
        self._read_event = asyncio.Event()
        self._sequence = 0

    async def headers_received(self, start_line, headers):
        await super(WSTerminalConnection, self).headers_received(start_line, headers)
        if start_line.code != 101:
            utils.logger.error(
                "[%s] Connect %s return %d"
                % (self.__class__.__name__, self._url, start_line.code)
            )
            self._connected = False
        else:
            self._connected = True

    async def wait_for_connecting(self):
        time0 = time.time()
        while time.time() - time0 < self.__timeout:
            if self._connected == None:
                await asyncio.sleep(0.005)
                continue
            if self._connected:
                asyncio.ensure_future(self.polling_packet_task())
                return True
            else:
                return False
        raise utils.ConnectWebsocketServerFailed("Connect %s timeout" % self._url)

    def on_message(self, message):
        if not message:
            self._closed = True
        else:
            self._buffer += message
            packet, self._buffer = proto.TransportPacket.deserialize(self._buffer)
            if packet:
                self._queue.put_nowait(packet.message)

    def on_connection_close(self):
        self._closed = True
        if self._handler:
            self._handler.on_connection_close()

    async def polling_packet_task(self):
        while not self._closed:
            if self._queue.empty():
                # Ensure the task exit in time
                await asyncio.sleep(0.005)
                continue
            packet = await self._queue.get()
            if packet["type"] == proto.EnumPacketType.REQUEST:
                await self.handle_request(packet)
            elif packet["type"] == proto.EnumPacketType.RESPONSE:
                await self.handle_response(packet)

    async def handle_request(self, request):
        if request["command"] == proto.EnumCommand.WRITE_STDOUT:
            sys.stdout.buffer.write(request["buffer"])
            sys.stdout.flush()
        elif request["command"] == proto.EnumCommand.SHELL_EXIT:
            if self._handler:
                self._handler.on_shell_exit()
        else:
            raise NotImplementedError(request["command"])

    async def handle_response(self, response):
        self._rsp_map[response["id"]] = response

    async def read_response(self, request, timeout=None):
        time0 = time.time()
        while not timeout or time.time() - time0 < timeout:
            if request["id"] in self._rsp_map:
                return self._rsp_map.pop(request["id"])
            await asyncio.sleep(0.005)
        return None

    async def send_request(self, command, **kwargs):
        self._sequence += 1
        data = {
            "command": command,
            "type": proto.EnumPacketType.REQUEST,
            "id": self._sequence,
        }
        data.update(kwargs)
        packet = proto.TransportPacket(data)
        await self.write_message(packet.serialize(), True)
        return data

    async def send_response(self, request, **kwargs):
        data = {
            "command": request["command"],
            "type": proto.EnumPacketType.RESPONSE,
            "id": request["id"],
        }
        data.update(kwargs)
        packet = proto.TransportPacket(data)
        return await self.write_message(packet.serialize(), True)


class WSTerminalClient(object):
    """Websocket Teminal Client"""

    file_fragment_size = 4 * 1024 * 1024

    def __init__(self, url, token=None, timeout=15, loop=None):
        headers = {}
        if token:
            headers["Authorization"] = "Token %s" % token
        self._conn = WSTerminalConnection(url, headers, timeout, handler=self)
        self._loop = loop or asyncio.get_event_loop()
        self._workspace = None
        self._running = True

    def on_connection_close(self):
        utils.logger.warn("[%s] Websocket connection closed" % self.__class__.__name__)
        self.on_shell_exit()

    async def connect(self):
        return await self._conn.wait_for_connecting()

    async def create_directory(self, dir_path):
        utils.logger.info(
            "[%s] Create directory %s" % (self.__class__.__name__, dir_path)
        )
        await self._conn.send_request(proto.EnumCommand.CREATE_DIR, path=dir_path)

    async def remove_directory(self, dir_path):
        utils.logger.info(
            "[%s] Remove directory %s" % (self.__class__.__name__, dir_path)
        )
        await self._conn.send_request(proto.EnumCommand.REMOVE_DIR, path=dir_path)

    async def update_workspace(self, dir_tree, root):
        assert "dirs" in dir_tree or "files" in dir_tree
        for name in dir_tree.get("dirs", {}):
            path = (root + "/" + name) if root else name
            if not dir_tree["dirs"][name]:
                # Blank directory
                await self.create_directory(path)
            elif dir_tree["dirs"][name] == "-":
                await self.remove_directory(path)
            else:
                await self.update_workspace(
                    dir_tree["dirs"][name], path,
                )
        for name in dir_tree.get("files", {}):
            path = (root + "/" + name) if root else name
            if dir_tree["files"][name] == "-":
                # Remove file
                await self.remove_file(path)
            else:
                await self.write_file(path)

    async def write_file(self, file_path):
        utils.logger.debug("[%s] Write file %s" % (self.__class__.__name__, file_path))
        offset = 0
        with open(self._workspace.join_path(file_path), "rb") as fp:
            while True:
                data = fp.read(self.file_fragment_size)
                if not data:
                    break
                await self._conn.send_request(
                    proto.EnumCommand.WRITE_FILE,
                    path=file_path,
                    data=data,
                    overwrite=offset == 0,
                )
                offset += len(data)

    async def remove_file(self, file_path):
        utils.logger.info("[%s] Remove file %s" % (self.__class__.__name__, file_path))
        await self._conn.send_request(proto.EnumCommand.REMOVE_FILE, path=file_path)

    async def move_item(self, src_path, dst_path):
        utils.logger.info(
            "[%s] Move item %s to %s" % (self.__class__.__name__, src_path, dst_path)
        )
        await self._conn.send_request(
            proto.EnumCommand.MOVE_ITEM, src_path=src_path, dst_path=dst_path
        )

    async def on_directory_created(self, path):
        await self.create_directory(path)

    async def on_directory_removed(self, path):
        await self.remove_directory(path)

    async def on_file_created(self, path):
        pass

    async def on_file_modified(self, path):
        await self.write_file(path)

    async def on_file_removed(self, path):
        await self.remove_file(path)

    async def on_item_moved(self, src_path, dst_path):
        await self.move_item(src_path, dst_path)

    async def sync_workspace(self, workspace_path):
        if not os.path.isdir(workspace_path):
            raise RuntimeError("Workspace %s not exist" % workspace_path)
        self._workspace = workspace.Workspace(workspace_path)
        workspace_hash = utils.make_short_hash(
            "%s%s" % (socket.gethostname(), workspace_path)
        )
        request = await self._conn.send_request(
            proto.EnumCommand.SYNC_WORKSPACE, workspace=workspace_hash,
        )
        response = await self._conn.read_response(request)
        diff_result = self._workspace.make_diff(response["data"])
        if diff_result:
            await self.update_workspace(diff_result, "")
        self._workspace.register_handler(self)
        asyncio.ensure_future(self._workspace.watch())

    async def write_shell_stdin(self, buffer):
        try:
            await self._conn.send_request(proto.EnumCommand.WRITE_STDIN, buffer=buffer)
        except tornado.websocket.WebSocketClosedError:
            utils.logger.error(
                "[%s] Websocket connection lost" % self.__class__.__name__
            )
            self.on_connection_close()

    def on_shell_exit(self):
        self._running = False

        async def exit_loop():
            await asyncio.sleep(0.01)
            self._loop.stop()

        asyncio.ensure_future(exit_loop())

    async def create_shell(self, size):
        request = await self._conn.send_request(
            proto.EnumCommand.CREATE_SHELL, size=size,
        )
        response = await self._conn.read_response(request)
        if response["code"]:
            raise RuntimeError("Create shell failed: %s" % response["message"])
        server_platform = response["platform"]
        if sys.platform != "win32":
            with utils.UnixStdIn() as shell_stdin:

                def on_input():
                    char = shell_stdin.read(1)
                    if char == b"\n":
                        char = b"\r"
                    if server_platform == "win32":
                        if char == b"\x1b":
                            chars = shell_stdin.read(2)
                            char += chars  # Must send together
                    utils.safe_ensure_future(self.write_shell_stdin(char))

                self._loop.add_reader(shell_stdin, on_input)

                while self._running:
                    await asyncio.sleep(0.005)
                self._loop.remove_reader(shell_stdin)
        else:
            import msvcrt

            while self._running:
                if msvcrt.kbhit():
                    char = msvcrt.getch()
                    if char == b"\xe0":
                        char = msvcrt.getch()
                        if char == b"H":
                            char = b"\x1b[A"
                        elif char == b"P":
                            char = b"\x1b[B"
                        elif char == b"K":
                            char = b"\x1b[D"
                        elif char == b"M":
                            char = b"\x1b[C"
                        else:
                            utils.logger.warn(
                                r"[%s] Unknown input key \xe0%s"
                                % (self.__class__.__name__, char)
                            )
                    elif char == b"\x1d":
                        # Ctrl + ]
                        char = b"\x03"
                    asyncio.ensure_future(self.write_shell_stdin(char))
                else:
                    await asyncio.sleep(0.005)



# -*- coding: utf-8 -*-


import argparse
import asyncio
import logging
import logging.handlers
import os
import shutil
import sys
import traceback
import urllib.parse

import tornado.ioloop

from . import client, server, utils


async def connect_server(url, workspace, token=None):
    cli = client.WSTerminalClient(url, token=token)
    if not await cli.connect():
        print(
            "Connect websocket server failed", file=sys.stderr
        )
        return False

    if workspace:
        print("Sync workspace to remote host...")
        await cli.sync_workspace(workspace)
        print("Sync workspace complete")
    terminal_size = shutil.get_terminal_size((120, 30))
    await cli.create_shell((terminal_size.columns, terminal_size.lines))
    return True


def main():
    parser = argparse.ArgumentParser(
        prog="wsterm", description="Websocket terminal tool."
    )
    parser.add_argument("--url", help="Websocket url", required=True)
    parser.add_argument(
        "--server", help="Run as websocket server", action="store_true", default=False
    )
    parser.add_argument("--token", help="Authorization token")
    parser.add_argument("--workspace", help="Workspace path")
    parser.add_argument(
        "--log-level",
        help="log level, default is info",
        choices=("debug", "info", "warn", "error"),
        default="info",
    )
    parser.add_argument("--log-file", help="Path to save log")
    parser.add_argument(
        "-d", "--daemon", help="Run as daemon", action="store_true", default=False
    )

    args = sys.argv[1:]
    if not args:
        parser.print_help()
        return 0

    args = parser.parse_args(args)

    url = urllib.parse.urlparse(args.url)
    if url.scheme != "ws":
        print("Error: Invalid websocket url %s" % args.url, file=sys.stderr)
        return -1

    log_file = None
    if args.log_file:
        log_file = os.path.abspath(args.log_file)

    if args.daemon:
        if not args.server:
            print("Error: -d/--daemon only supported on server", file=sys.stderr)
            return -1

        if sys.platform != "win32":
            import daemon

            daemon.DaemonContext(stderr=open("error.txt", "w")).open()
        else:
            utils.win32_daemon()
            return 0

    handler = logging.StreamHandler()
    formatter = logging.Formatter("[%(asctime)s][%(levelname)s]%(message)s")
    handler.setFormatter(formatter)

    if args.log_level == "debug":
        utils.logger.setLevel(logging.DEBUG)
    elif args.log_level == "info":
        utils.logger.setLevel(logging.INFO)
    elif args.log_level == "warn":
        utils.logger.setLevel(logging.WARN)
    elif args.log_level == "error":
        utils.logger.setLevel(logging.ERROR)

    utils.logger.propagate = 0
    if args.server:
        utils.logger.addHandler(handler)
    else:
        log_file = log_file or "wsterm.log"

    if log_file:
        handler = logging.handlers.RotatingFileHandler(
            log_file, maxBytes=10 * 1024 * 1024, backupCount=4
        )
        formatter = logging.Formatter(
            "[%(asctime)s][%(levelname)s][%(filename)s][%(lineno)d]%(message)s"
        )
        handler.setFormatter(formatter)
        utils.logger.addHandler(handler)

    if args.server:
        host = url.hostname
        port = url.port or 80
        if sys.platform == "win32":
            loop = asyncio.ProactorEventLoop()
            asyncio.set_event_loop(loop)
        server.start_server((host, port), url.path, args.token)
    else:
        if sys.platform == "win32":
            utils.enable_native_ansi()

        if not asyncio.get_event_loop().run_until_complete(
            connect_server(args.url, args.workspace, args.token)
        ):
            return -1

    def handle_exception(loop, context):
        print("Exception caught:\n", file=sys.stderr)
        message = context["message"]
        exp = context.get("exception")
        if exp:
            message = "".join(
                traceback.format_exception(
                    etype=type(exp), value=exp, tb=exp.__traceback__
                )
            )
        print(message, file=sys.stderr)
        loop.stop()

    loop = asyncio.get_event_loop()
    loop.set_exception_handler(handle_exception)

    try:
        tornado.ioloop.IOLoop.current().start()
    except KeyboardInterrupt:
        print("Process exit warmly.")


if __name__ == "__main__":
    sys.exit(main())



# -*- coding: utf-8 -*-

import asyncio
import os
import select

from . import WatcherBackendBase, WatchEvent


O_EVTONLY = 0x8000


class KEventsWatcher(WatcherBackendBase):
    def __init__(self, loop=None):
        super(KEventsWatcher, self).__init__(loop)
        self._kq = select.kqueue()
        self._watch_list = {}
        self._dir_tree = {}
        asyncio.ensure_future(self.polling_task())

    def _get_dir_node(self, path, auto_create=False):
        if path[0] == "/":
            path = path[1:]
        items = path.split("/")
        node = self._dir_tree
        for it in items:
            if it not in node:
                if auto_create:
                    node[it] = {}
                else:
                    return None
            node = node[it]
        return node

    def _get_dir_new_item(self, path):
        result = []
        node = self._get_dir_node(path, True)
        items = os.listdir(path)
        for it in items:
            if it not in node:
                subpath = os.path.join(path, it)
                if os.path.isdir(subpath):
                    node[it] = {}
                    result.append(WatchEvent(WatchEvent.DIRECTORY_CREATED, subpath))
                    result.extend(self._get_dir_new_item(subpath))
                elif os.path.isfile(subpath):
                    node[it] = None
                    result.append(WatchEvent(WatchEvent.FILE_CREATED, subpath))

        return result

    def add_dir_watch(self, path):
        assert os.path.isdir(path)
        self.add_watch(path)
        node = self._get_dir_node(path, True)
        for it in os.listdir(path):
            subpath = os.path.join(path, it)
            if os.path.isdir(subpath):
                node[it] = {}
                self.add_dir_watch(subpath)
            elif os.path.isfile(subpath):
                node[it] = None
                self.add_watch(subpath)

    def add_watch(self, path):
        fd = os.open(path, O_EVTONLY)
        event = select.kevent(
            fd,
            filter=select.KQ_FILTER_VNODE,
            flags=select.KQ_EV_ADD | select.KQ_EV_ENABLE | select.KQ_EV_CLEAR,
            fflags=select.KQ_NOTE_DELETE
            | select.KQ_NOTE_WRITE
            | select.KQ_NOTE_EXTEND
            | select.KQ_NOTE_ATTRIB
            | select.KQ_NOTE_LINK
            | select.KQ_NOTE_RENAME
            | select.KQ_NOTE_REVOKE,
        )
        self._watch_list[event.ident] = (path, event, fd)

    def remove_watch(self, path):
        for evt_id in self._watch_list:
            watch_path, _, fd = self._watch_list[evt_id]
            if path == watch_path:
                event = select.kevent(
                    fd,
                    filter=select.KQ_FILTER_VNODE,
                    flags=select.KQ_EV_DELETE,
                )
                # self._kq.control((event,), 0)
                self._watch_list.pop(evt_id)
                return

    async def polling_task(self):
        while True:
            events = [event for _, event, _ in self._watch_list.values()]
            events = list(self._kq.control(events, 4096, 0))
            if not events:
                await asyncio.sleep(0.005)
                continue
            events.sort(
                key=lambda event: len(self._watch_list[event.ident][0].split("/")),
                reverse=True,
            )  # Ensure remove inner items first when remove directory

            for event in events:
                target, _, _ = self._watch_list[event.ident]
                if not os.path.exists(target):
                    # Item removed
                    path = target
                    remove_root = target
                    while not os.path.exists(path):
                        # Get remove root
                        remove_root = path
                        path = os.path.dirname(path)

                    if self._get_dir_node(remove_root) is not None:
                        # Insert remove sub dirs/files event
                        def _handle_sub_dir(path):
                            dir_node = self._get_dir_node(path)
                            for it in dir_node:
                                if isinstance(dir_node[it], dict):
                                    if dir_node[it]:
                                        _handle_sub_dir(os.path.join(path, it))
                                    self._event_queue.put_nowait(
                                        WatchEvent(
                                            WatchEvent.DIRECTORY_REMOVED,
                                            os.path.join(path, it),
                                        )
                                    )
                                else:
                                    self._event_queue.put_nowait(
                                        WatchEvent(
                                            WatchEvent.FILE_REMOVED,
                                            os.path.join(path, it),
                                        )
                                    )

                        _handle_sub_dir(remove_root)
                        self._event_queue.put_nowait(
                            WatchEvent(WatchEvent.DIRECTORY_REMOVED, remove_root)
                        )
                    else:
                        self._event_queue.put_nowait(
                            WatchEvent(WatchEvent.FILE_REMOVED, remove_root)
                        )

                    parent = self._get_dir_node(os.path.dirname(target))
                    parent.pop(os.path.split(target)[-1])
                    self.remove_watch(target)
                    continue
                elif os.path.isdir(target):
                    new_events = self._get_dir_new_item(target)
                    for event in new_events:
                        self._event_queue.put_nowait(event)
                        if event.event == WatchEvent.FILE_CREATED:
                            self.add_watch(event.target)
                            self._event_queue.put_nowait(
                                WatchEvent(WatchEvent.FILE_MODIFIED, event.target)
                            )
                        elif event.event == WatchEvent.DIRECTORY_CREATED:
                            self.add_dir_watch(event.target)
                    continue
                else:
                    event = WatchEvent(WatchEvent.FILE_MODIFIED, target)

                self._event_queue.put_nowait(event)

    async def read_event(self):
        return await self._event_queue.get()



# -*- coding: utf-8 -*-

import asyncio
import ctypes
import ctypes.util
import errno
import functools
import os
import struct

from . import WatcherBackendBase, WatchEvent


def _load_libc():
    libc_path = None
    try:
        libc_path = ctypes.util.find_library("c")
    except (OSError, RuntimeError):
        # Note: find_library will on some platforms raise these undocumented
        # errors, e.g.on android OSError "No usable temporary directory found"
        # will be raised.
        pass

    if libc_path is not None:
        return ctypes.CDLL(libc_path)

    # Fallbacks
    try:
        return ctypes.CDLL("libc.so")
    except OSError:
        pass

    try:
        return ctypes.CDLL("libc.so.6")
    except OSError:
        pass

    # uClibc
    try:
        return ctypes.CDLL("libc.so.0")
    except OSError as err:
        raise err


libc = _load_libc()

if (
    not hasattr(libc, "inotify_init")
    or not hasattr(libc, "inotify_add_watch")
    or not hasattr(libc, "inotify_rm_watch")
):
    raise RuntimeError("Unsupported libc version found: %s" % libc._name)


class InotifyConstants(object):
    # User-space events
    IN_ACCESS = 0x00000001  # File was accessed.
    IN_MODIFY = 0x00000002  # File was modified.
    IN_ATTRIB = 0x00000004  # Meta-data changed.
    IN_CLOSE_WRITE = 0x00000008  # Writable file was closed.
    IN_CLOSE_NOWRITE = 0x00000010  # Unwritable file closed.
    IN_OPEN = 0x00000020  # File was opened.
    IN_MOVED_FROM = 0x00000040  # File was moved from X.
    IN_MOVED_TO = 0x00000080  # File was moved to Y.
    IN_CREATE = 0x00000100  # Subfile was created.
    IN_DELETE = 0x00000200  # Subfile was deleted.
    IN_DELETE_SELF = 0x00000400  # Self was deleted.
    IN_MOVE_SELF = 0x00000800  # Self was moved.

    # Helper user-space events.
    IN_CLOSE = IN_CLOSE_WRITE | IN_CLOSE_NOWRITE  # Close.
    IN_MOVE = IN_MOVED_FROM | IN_MOVED_TO  # Moves.

    # Events sent by the kernel to a watch.
    IN_UNMOUNT = 0x00002000  # Backing file system was unmounted.
    IN_Q_OVERFLOW = 0x00004000  # Event queued overflowed.
    IN_IGNORED = 0x00008000  # File was ignored.

    # Special flags.
    IN_ONLYDIR = 0x01000000  # Only watch the path if it's a directory.
    IN_DONT_FOLLOW = 0x02000000  # Do not follow a symbolic link.
    IN_EXCL_UNLINK = 0x04000000  # Exclude events on unlinked objects
    IN_MASK_ADD = 0x20000000  # Add to the mask of an existing watch.
    IN_ISDIR = 0x40000000  # Event occurred against directory.
    IN_ONESHOT = 0x80000000  # Only send event once.

    # All user-space events.
    IN_ALL_EVENTS = functools.reduce(
        lambda x, y: x | y,
        [
            IN_ACCESS,
            IN_MODIFY,
            IN_ATTRIB,
            IN_CLOSE_WRITE,
            IN_CLOSE_NOWRITE,
            IN_OPEN,
            IN_MOVED_FROM,
            IN_MOVED_TO,
            IN_DELETE,
            IN_CREATE,
            IN_DELETE_SELF,
            IN_MOVE_SELF,
        ],
    )

    # Flags for ``inotify_init1``
    IN_CLOEXEC = 0x02000000
    IN_NONBLOCK = 0x00004000

    @staticmethod
    def parse(mask):
        result = []
        for key in dir(InotifyConstants):
            if not key.startswith("IN_"):
                continue
            if mask & getattr(InotifyConstants, key):
                result.append(key)
        return result


class inotify_event_struct(ctypes.Structure):
    """
    Structure representation of the inotify_event structure
    (used in buffer size calculations)::

        struct inotify_event {
            __s32 wd;            /* watch descriptor */
            __u32 mask;          /* watch mask */
            __u32 cookie;        /* cookie to synchronize two events */
            __u32 len;           /* length (including nulls) of name */
            char  name[0];       /* stub for possible name */
        };
    """

    _fields_ = [
        ("wd", ctypes.c_int),
        ("mask", ctypes.c_uint32),
        ("cookie", ctypes.c_uint32),
        ("len", ctypes.c_uint32),
        ("name", ctypes.c_char_p),
    ]


EVENT_SIZE = ctypes.sizeof(inotify_event_struct)
DEFAULT_NUM_EVENTS = 2048
DEFAULT_EVENT_BUFFER_SIZE = DEFAULT_NUM_EVENTS * (EVENT_SIZE + 16)


class INotifyWatcher(WatcherBackendBase):
    """inotify implementation"""

    def __init__(self, loop=None):
        super(INotifyWatcher, self).__init__(loop)
        self._inotify_fd = libc.inotify_init()
        if self._inotify_fd == -1:
            INotifyWatcher._raise_error()
        self._loop.add_reader(self._inotify_fd, self._read_event)
        self._watch_list = {}

    async def read_event(self):
        move_from = None
        while True:
            target, mask = await self._event_queue.get()
            isdir = mask & InotifyConstants.IN_ISDIR
            if mask & InotifyConstants.IN_CREATE:
                if isdir:
                    return WatchEvent(WatchEvent.DIRECTORY_CREATED, target)
                else:
                    return WatchEvent(WatchEvent.FILE_CREATED, target)
            elif mask & InotifyConstants.IN_MODIFY:
                if not isdir:
                    return WatchEvent(WatchEvent.FILE_MODIFIED, target)
            elif mask & InotifyConstants.IN_DELETE:
                if isdir:
                    return WatchEvent(WatchEvent.DIRECTORY_REMOVED, target)
                else:
                    return WatchEvent(WatchEvent.FILE_REMOVED, target)
            elif (
                mask & InotifyConstants.IN_MOVE
                and mask & InotifyConstants.IN_MOVED_FROM
            ):
                move_from = target
            elif (
                mask & InotifyConstants.IN_MOVE and mask & InotifyConstants.IN_MOVED_TO
            ):
                return WatchEvent(WatchEvent.ITEM_MOVED, (move_from, target))

    def add_dir_watch(self, path, mask=InotifyConstants.IN_ALL_EVENTS):
        assert os.path.isdir(path)
        for it in os.listdir(path):
            subpath = os.path.join(path, it)
            if os.path.isdir(subpath):
                self.add_dir_watch(subpath, mask)
        self.add_watch(path, mask)

    def add_watch(self, path, mask=InotifyConstants.IN_ALL_EVENTS):
        wd = libc.inotify_add_watch(self._inotify_fd, path.encode(), mask)
        if wd == -1:
            INotifyWatcher._raise_error()
        self._watch_list[wd] = path

    def _read_event(self, event_buffer_size=DEFAULT_EVENT_BUFFER_SIZE):
        event_buffer = os.read(self._inotify_fd, event_buffer_size)
        i = 0
        while i + 16 <= len(event_buffer):
            wd, mask, _, length = struct.unpack_from("iIII", event_buffer, i)
            name = event_buffer[i + 16 : i + 16 + length].rstrip(b"\0")
            i += 16 + length
            target = self._watch_list.get(wd)
            if name:
                target = os.path.join(target, name.decode())

            self._event_queue.put_nowait((target, mask))
            if mask & InotifyConstants.IN_CREATE and mask & InotifyConstants.IN_ISDIR:
                # Handle mkdir -p
                def _handle_sub_dir(path):
                    for it in os.listdir(path):
                        mask = InotifyConstants.IN_CREATE
                        subpath = os.path.join(path, it)
                        if os.path.isdir(subpath):
                            mask |= InotifyConstants.IN_ISDIR
                        self._event_queue.put_nowait((subpath, mask))
                        if os.path.isdir(subpath):
                            _handle_sub_dir(subpath)

                _handle_sub_dir(target)
                self.add_dir_watch(target)

    @staticmethod
    def _raise_error():
        """
        Raises errors for inotify failures.
        """
        err = ctypes.get_errno()
        if err == errno.ENOSPC:
            raise OSError(errno.ENOSPC, "inotify watch limit reached")
        elif err == errno.EMFILE:
            raise OSError(errno.EMFILE, "inotify instance limit reached")
        elif err == errno.EACCES:
            # Prevent raising an exception when a file with no permissions
            # changes
            pass
        else:
            raise OSError(err, os.strerror(err))



# -*- coding: utf-8 -*-

import asyncio
import ctypes
import os
import shlex
import struct
import sys

from . import utils


class Shell(object):
    def __init__(self, workspace, size=None):
        self._workspace = workspace
        self._size = size or (80, 23)
        self._proc = None
        self._fd = None
        self._stdin = None
        self._stdout = None
        self._stderr = None

    @property
    def stdin(self):
        return self._stdin

    @property
    def stdout(self):
        return self._stdout

    @property
    def stderr(self):
        return self._stderr

    async def create(self):
        if sys.platform == "win32":
            if hasattr(ctypes.windll.kernel32, "CreatePseudoConsole"):
                cmd = (
                    "conhost.exe",
                    "--headless",
                    "--width",
                    str(self._size[0]),
                    "--height",
                    str(self._size[1]),
                    "--",
                    "cmd.exe",
                )
            else:
                cmd = ("cmd.exe",)
            proc = await asyncio.create_subprocess_exec(
                *cmd,
                cwd=self._workspace,
                stdin=asyncio.subprocess.PIPE,
                stdout=asyncio.subprocess.PIPE,
                stderr=asyncio.subprocess.PIPE,
                close_fds=False
            )
            self._stdin = proc.stdin
            self._stdout = proc.stdout
            self._stderr = proc.stderr
        else:
            import pty

            cmdline = list(shlex.split(os.environ.get("SHELL", "sh")))
            exe = cmdline[0]
            if exe[0] != "/":
                for it in os.environ["PATH"].split(":"):
                    path = os.path.join(it, exe)
                    if os.path.isfile(path):
                        exe = path
                        break

            utils.logger.info(
                "[%s] Create shell %s" % (self.__class__.__name__, cmdline)
            )
            pid, self._fd = pty.fork()
            if pid == 0:
                # child process
                sys.stdout.flush()
                os.chdir(self._workspace)
                try:
                    os.execve(exe, cmdline, os.environ)
                except Exception as e:
                    sys.stderr.write(str(e))
            else:
                proc = utils.Process(pid)
                self._stdin = utils.AsyncFileDescriptor(self._fd)
                self._stdout = utils.AsyncFileDescriptor(self._fd)
                self.resize(self._size)

        return proc, self._stdin, self._stdout, self._stderr

    def write(self, buffer):
        self._stdin.write(buffer)

    def resize(self, size):
        if sys.platform == "win32":
            pass
        else:
            import fcntl
            import termios

            winsize = struct.pack("HHHH", size[1], size[0], 0, 0)
            fcntl.ioctl(self._fd, termios.TIOCSWINSZ, winsize)
        self._size = size
        return True



# -*- coding: utf-8 -*-

import setuptools

import wsterm

with open("README.md", "rb") as fp:
    README = fp.read().decode()


with open("requirements.txt") as fp:
    text = fp.read()
    REQUIREMENTS = text.split("\n")


def find_packages():
    packages = []
    for pkg in setuptools.find_packages():
        if not pkg.startswith("test"):
            print(pkg)
            packages.append(pkg)
    return packages


setuptools.setup(
    author="drunkdream",
    author_email="drunkdream@qq.com",
    name="wsterm",
    license="MIT",
    description="Websocket remote debug terminal (auto sync workspace).",
    version=wsterm.VERSION,
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://github.com/wsterm/wsterm",
    packages=find_packages(),
    python_requires=">=3.5",
    install_requires=REQUIREMENTS,
    extras_require={},
    classifiers=[
        # Trove classifiers
        # (https://pypi.python.org/pypi?%3Aaction=list_classifiers)
        "Development Status :: 3 - Alpha",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Intended Audience :: Developers",
    ],
    entry_points={
        "console_scripts": [
            "wsterm = wsterm.__main__:main",
        ],
    },
)



# -*- coding: utf-8 -*-

import asyncio
import hashlib
import os
import shutil

import gitignore_parser

from . import aiowatch, utils


class EnumEvent(object):
    """Workspace event"""

    # BEFORE_SNAPSHOT = "before_snapshot"
    # AFTER_SNAPSHOT = "after_snapshot"
    ON_DIRECTORY_CREATED = "on_directory_created"
    ON_DIRECTORY_REMOVED = "on_directory_removed"
    ON_FILE_CREATED = "on_file_created"
    ON_FILE_REMOVED = "on_file_removed"
    ON_FILE_MODIFIED = "on_file_modified"
    ON_ITEM_MOVED = "on_item_moved"


class Directory(object):
    def __init__(self, dir_path):
        self._dir_path = dir_path

    @property
    def name(self):
        return os.path.split(self._dir_path)[-1]

    @property
    def path(self):
        return self._dir_path

    def get_dirs(self):
        dir_list = []
        for it in os.listdir(self._dir_path):
            path = os.path.join(self._dir_path, it)
            if os.path.isdir(path):
                dir_list.append(Directory(path))
        return dir_list

    def get_files(self):
        file_list = []
        for it in os.listdir(self._dir_path):
            path = os.path.join(self._dir_path, it)
            if os.path.isfile(path):
                file_list.append(File(path))
        return file_list


class File(object):
    def __init__(self, file_path):
        self._file_path = file_path

    @property
    def name(self):
        return os.path.split(self._file_path)[-1]

    @property
    def path(self):
        return self._file_path

    @property
    def last_modify_time(self):
        return os.stat(self._file_path).st_mtime

    @property
    def size(self):
        return os.path.getsize(self._file_path)

    @property
    def hash(self):
        m = hashlib.md5()
        with open(self._file_path, "rb") as fp:
            m.update(fp.read())
            return m.hexdigest()


class Workspace(object):
    def __init__(self, root_path):
        self._root_path = root_path
        if not os.path.isdir(self._root_path):
            os.makedirs(self._root_path)
        self._handlers = []
        self._watcher = aiowatch.AIOWatcher(self._root_path, self)
        self._running = True

    @property
    def path(self):
        return self._root_path

    def join_path(self, path):
        return os.path.join(self._root_path, path.replace("/", os.path.sep))

    def create_directory(self, path):
        dir_path = os.path.join(self._root_path, path.replace("/", os.path.sep))
        if not os.path.isdir(dir_path):
            os.makedirs(dir_path)

    def remove_directory(self, path):
        dir_path = os.path.join(self._root_path, path.replace("/", os.path.sep))
        if os.path.isdir(dir_path):
            shutil.rmtree(dir_path)

    def write_file(self, path, data, overwrite=True):
        file_path = os.path.join(self._root_path, path.replace("/", os.path.sep))
        dir_path = os.path.dirname(file_path)
        if not os.path.isdir(dir_path):
            os.makedirs(dir_path)
        flag = "wb" if overwrite else "ab"
        with open(file_path, flag) as fp:
            fp.write(data)

    def remove_file(self, path):
        file_path = os.path.join(self._root_path, path.replace("/", os.path.sep))
        if os.path.isfile(file_path):
            os.remove(file_path)

    def move_item(self, src_path, dst_path):
        src_path = os.path.join(self._root_path, src_path.replace("/", os.path.sep))
        dst_path = os.path.join(self._root_path, dst_path.replace("/", os.path.sep))
        os.rename(src_path, dst_path)

    def register_handler(self, handler):
        self._handlers.append(handler)

    def on_event(self, event_name, **kwargs):
        path = kwargs.get("path", kwargs.get("src_path"))
        if path and ".git" in path.split(os.path.sep):
            return
        for handler in self._handlers:
            func = getattr(handler, event_name)
            if func:
                asyncio.ensure_future(func(**kwargs))

    def on_directory_created(self, path):
        self.on_event(EnumEvent.ON_DIRECTORY_CREATED, path=path)

    def on_directory_removed(self, path):
        self.on_event(EnumEvent.ON_DIRECTORY_REMOVED, path=path)

    def on_file_created(self, path):
        self.on_event(EnumEvent.ON_FILE_CREATED, path=path)

    def on_file_modified(self, path):
        self.on_event(EnumEvent.ON_FILE_MODIFIED, path=path)

    def on_file_removed(self, path):
        self.on_event(EnumEvent.ON_FILE_REMOVED, path=path)

    def on_item_moved(self, src_path, dst_path):
        self.on_event(EnumEvent.ON_ITEM_MOVED, src_path=src_path, dst_path=dst_path)

    def snapshot(self, root=None):
        result = {"dirs": {}, "files": {}}
        root = root or self._root_path
        if os.path.isdir(root) and os.path.split(root)[-1] == ".git":
            # Auto ignore .git directory
            return
        gitignore_path = os.path.join(self._root_path, ".gitignore")
        if os.path.isfile(gitignore_path):
            matches = gitignore_parser.parse_gitignore(gitignore_path)
            match = False
            try:
                match = matches(root)
            except ValueError:
                utils.logger.info(
                    "[%s] Ignore invalid path %s" % (self.__class__.__name__, root)
                )
                return None

            if match:
                utils.logger.debug(
                    "[%s] Path %s ignored" % (self.__class__.__name__, root)
                )
                return None

        root_dir = Directory(root)
        for subdir in root_dir.get_dirs():
            res = self.snapshot(subdir.path)
            if res:
                result["dirs"][subdir.name] = res
        for file in root_dir.get_files():
            if file.name.endswith(".pyc"):
                # Auto ignore .pyc files
                continue
            result["files"][file.name] = file.hash
        return result

    async def watch(self):
        await self._watcher.start()

    def make_diff(self, other_snapshot):
        cache_data = self.snapshot()
        return utils.diff(cache_data, other_snapshot)



# -*- coding: utf-8 -*-

"""Protocol
"""

import struct

import msgpack


class EnumPacketType(object):
    REQUEST = 1
    RESPONSE = 2


class EnumCommand(object):
    SYNC_WORKSPACE = "sync-workspace"
    LIST_DIR = "list-dir"
    CREATE_DIR = "create-dir"
    REMOVE_DIR = "remove-dir"
    WRITE_FILE = "write-file"
    REMOVE_FILE = "remove-file"
    MOVE_ITEM = "move-item"

    CREATE_SHELL = "create-shell"
    WRITE_STDIN = "write-stdin"
    WRITE_STDOUT = "write-stdout"
    WRITE_STDERR = "write-stderr"
    SHELL_EXIT = "shell-exit"


class TransportPacket(object):
    """Transport packet serialize/deserialize"""

    def __init__(self, message):
        self._message = message

    def __str__(self):
        return "<%s object command=%s type=%d at 0x%x>" % (
            self.__class__.__name__,
            self._message.get("command"),
            self._message.get("type"),
            id(self),
        )

    @property
    def message(self):
        return self._message

    def serialize(self):
        buffer = msgpack.dumps(self._message)
        return struct.pack("!I", len(buffer)) + buffer

    @staticmethod
    def deserialize(buffer):
        if len(buffer) < 4:
            return None, buffer
        buffer_size = struct.unpack("!I", buffer[:4])[0]
        if len(buffer) - 4 < buffer_size:
            return None, buffer
        message = msgpack.loads(buffer[4 : 4 + buffer_size])
        buffer = buffer[4 + buffer_size :]
        return TransportPacket(message), buffer



# -*- coding: utf-8 -*-

import asyncio
import ctypes
import os

import pywintypes
import win32con
import win32event
import win32file

from . import WatcherBackendBase, WatchEvent

FILE_LIST_DIRECTORY = 1
FILE_ACTION_ADDED = 1
FILE_ACTION_REMOVED = 2
FILE_ACTION_MODIFIED = 3
FILE_ACTION_RENAMED_OLD_NAME = 4
FILE_ACTION_RENAMED_NEW_NAME = 5


class EnumWatchType(object):

    WATCH_DIRECTORY = 1
    WATCH_FILE = 2


class Win32Watcher(WatcherBackendBase):
    def __init__(self, loop=None):
        super(Win32Watcher, self).__init__(loop)
        self._watch_list = []
        asyncio.ensure_future(self.polling_task())
        self._dir_tree = {}

    def _get_dir_node(self, path, auto_create=False):
        path = path.replace(":\\\\", "\\")
        items = path.split("\\")
        node = self._dir_tree
        for it in items:
            if it not in node:
                if auto_create:
                    node[it] = {}
                else:
                    return None
            node = node[it]
        return node

    def _snapshot(self, path):
        assert os.path.isdir(path)
        node = self._get_dir_node(path, True)
        items = os.listdir(path)
        for it in items:
            if it not in node:
                subpath = os.path.join(path, it)
                if os.path.isdir(subpath):
                    node[it] = {}
                    self._snapshot(subpath)
                elif os.path.isfile(subpath):
                    node[it] = None

    def _create_watch_handle(self, path):
        return win32file.CreateFile(
            path,
            FILE_LIST_DIRECTORY,
            win32con.FILE_SHARE_READ | win32con.FILE_SHARE_WRITE,
            None,
            win32con.OPEN_EXISTING,
            win32con.FILE_FLAG_BACKUP_SEMANTICS | win32con.FILE_FLAG_OVERLAPPED,
            None,
        )

    def _add_file_watch(self, handle, buffer, ov):
        win32file.ReadDirectoryChangesW(
            handle,
            buffer,
            True,
            win32con.FILE_NOTIFY_CHANGE_FILE_NAME
            | win32con.FILE_NOTIFY_CHANGE_ATTRIBUTES
            | win32con.FILE_NOTIFY_CHANGE_SIZE
            | win32con.FILE_NOTIFY_CHANGE_LAST_WRITE
            | win32con.FILE_NOTIFY_CHANGE_SECURITY,
            ov,
        )

    def _add_dir_watch(self, handle, buffer, ov):
        win32file.ReadDirectoryChangesW(
            handle, buffer, True, win32con.FILE_NOTIFY_CHANGE_DIR_NAME, ov,
        )

    def add_dir_watch(self, path):
        # Add directory watch
        self._snapshot(path)
        handle_dir = self._create_watch_handle(path)
        ov_dir = pywintypes.OVERLAPPED()
        ov_dir.hEvent = win32event.CreateEvent(None, 0, 0, None)
        buffer_dir = win32file.AllocateReadBuffer(8192)
        self._add_dir_watch(handle_dir, buffer_dir, ov_dir)
        self._watch_list.append(
            (EnumWatchType.WATCH_DIRECTORY, path, handle_dir, ov_dir, buffer_dir)
        )

        # Add file watch
        handle_file = self._create_watch_handle(path)
        ov_file = pywintypes.OVERLAPPED()
        ov_file.hEvent = win32event.CreateEvent(None, 0, 0, None)
        buffer_file = win32file.AllocateReadBuffer(8192)
        self._add_file_watch(handle_file, buffer_file, ov_file)
        self._watch_list.append(
            (EnumWatchType.WATCH_FILE, path, handle_file, ov_file, buffer_file)
        )

    async def polling_task(self):
        while True:
            for item in self._watch_list:
                try:
                    length = win32file.GetOverlappedResult(item[2], item[3], 0)
                except pywintypes.error as e:
                    if e.args[0] == 996:
                        # Overlapped I/O event is not in a signaled state
                        continue
                    else:
                        raise e
                else:
                    for action, name in win32file.FILE_NOTIFY_INFORMATION(
                        item[-1], length
                    ):
                        target = os.path.join(item[1], name)
                        if action == FILE_ACTION_MODIFIED and not os.path.isfile(
                            target
                        ):
                            # Ignore directory modify event
                            continue
                        elif action == FILE_ACTION_MODIFIED and os.path.isfile(target):
                            dir_node = self._get_dir_node(item[1])
                            if name not in dir_node:
                                # Auto insert file create event
                                self._event_queue.put_nowait(
                                    (target, item[0], FILE_ACTION_ADDED)
                                )
                                dir_node[name] = None
                        elif action == FILE_ACTION_REMOVED:
                            path = target
                            remove_root = target
                            while not os.path.exists(path):
                                # Get remove root
                                remove_root = path
                                path = os.path.dirname(path)

                            if self._get_dir_node(remove_root):
                                # Insert remove sub dirs/files event
                                def _handle_sub_dir(path):
                                    dir_node = self._get_dir_node(path)
                                    for it in dir_node:
                                        if isinstance(dir_node[it], dict):
                                            if dir_node[it]:
                                                _handle_sub_dir(os.path.join(path, it))
                                            self._event_queue.put_nowait(
                                                (
                                                    os.path.join(path, it),
                                                    EnumWatchType.WATCH_DIRECTORY,
                                                    action,
                                                )
                                            )
                                        else:
                                            self._event_queue.put_nowait(
                                                (
                                                    os.path.join(path, it),
                                                    EnumWatchType.WATCH_FILE,
                                                    action,
                                                )
                                            )

                                _handle_sub_dir(remove_root)
                                self._event_queue.put_nowait(
                                    (
                                        remove_root,
                                        EnumWatchType.WATCH_DIRECTORY,
                                        action,
                                    )
                                )
                                parent_dir_node = self._get_dir_node(
                                    os.path.dirname(remove_root)
                                )
                                parent_dir_node.pop(os.path.split(remove_root)[-1])
                                continue

                        self._event_queue.put_nowait((target, item[0], action))
                        if action == FILE_ACTION_ADDED and os.path.isdir(target):
                            # Handle mkdir -p
                            def _handle_sub_dir(path):
                                dir_node = self._get_dir_node(path, True)
                                for it in os.listdir(path):
                                    if it in dir_node:
                                        continue
                                    subpath = os.path.join(path, it)
                                    if os.path.isdir(subpath):
                                        watch_type = EnumWatchType.WATCH_DIRECTORY
                                        dir_node[it] = {}
                                    else:
                                        watch_type = EnumWatchType.WATCH_FILE
                                        dir_node[it] = None

                                    self._event_queue.put_nowait(
                                        (subpath, watch_type, action)
                                    )

                                    if os.path.isdir(subpath):
                                        _handle_sub_dir(subpath)

                            dir_node = self._get_dir_node(item[1])
                            dir_node[name] = {}
                            _handle_sub_dir(target)
                        elif action == FILE_ACTION_ADDED and os.path.isfile(target):
                            # Auto fire modify event
                            self._event_queue.put_nowait(
                                (target, item[0], FILE_ACTION_MODIFIED)
                            )
                            dir_node = self._get_dir_node(os.path.dirname(target))
                            dir_node[os.path.split(target)[-1]] = None

                    # Continue to listen
                    self._add_dir_watch(item[2], item[4], item[3])
                    self._add_file_watch(item[2], item[4], item[3])

            await asyncio.sleep(0.005)

    async def read_event(self):
        move_from = None
        while True:
            target, watch_type, action = await self._event_queue.get()
            isdir = watch_type == EnumWatchType.WATCH_DIRECTORY
            if action == FILE_ACTION_ADDED:
                if isdir:
                    return WatchEvent(WatchEvent.DIRECTORY_CREATED, target)
                else:
                    return WatchEvent(WatchEvent.FILE_CREATED, target)
            elif action == FILE_ACTION_MODIFIED:
                if not isdir:
                    return WatchEvent(WatchEvent.FILE_MODIFIED, target)
            elif action == FILE_ACTION_REMOVED:
                if isdir:
                    return WatchEvent(WatchEvent.DIRECTORY_REMOVED, target)
                else:
                    return WatchEvent(WatchEvent.FILE_REMOVED, target)
            elif action == FILE_ACTION_RENAMED_OLD_NAME:
                move_from = target
            elif action == FILE_ACTION_RENAMED_NEW_NAME:
                return WatchEvent(WatchEvent.ITEM_MOVED, (move_from, target))



# -*- coding: utf-8 -*-

"""Watch File System Asynchronously
"""

import asyncio
import sys


class WatcherBackendBase(object):
    def __init__(self, loop=None):
        self._loop = loop or asyncio.get_event_loop()
        self._event_queue = asyncio.Queue()

    def add_dir_watch(self, path):
        raise NotImplementedError()

    def add_watch(self, path):
        raise NotImplementedError()

    async def read_event(self):
        raise NotImplementedError()


class WatchEvent(object):
    DIRECTORY_CREATED = 0
    DIRECTORY_REMOVED = 1
    FILE_CREATED = 2
    FILE_MODIFIED = 3
    FILE_REMOVED = 4
    ITEM_MOVED = 5

    def __init__(self, event, target):
        self._event = event
        self._target = target

    def __str__(self):
        return "<%s object event=%s target=%s at 0x%x>" % (
            self.__class__.__name__,
            self._event,
            self._target,
            id(self),
        )

    @property
    def event(self):
        return self._event

    @property
    def target(self):
        return self._target


class AIOWatcher(object):
    def __init__(self, root_path, handler):
        self._root_path = root_path
        self._backend = self._get_backend()
        self._handler = handler
        self._running = True

    def _get_backend(self):
        if sys.platform == "linux":
            from . import inotify

            return inotify.INotifyWatcher()
        elif sys.platform == "win32":
            from . import win32

            return win32.Win32Watcher()
        elif sys.platform == "darwin":
            from . import kevents

            return kevents.KEventsWatcher()
        else:
            raise NotImplementedError(sys.platform)

    def add_dir_watch(self, path):
        self._backend.add_dir_watch(path)

    def add_watch(self, path):
        self._backend.add_watch(path)

    async def start(self):
        self.add_dir_watch(self._root_path)
        while self._running:
            event = await self._backend.read_event()
            if isinstance(event.target, str):
                target = event.target[len(self._root_path) + 1 :]
            else:
                target = (
                    event.target[0][len(self._root_path) + 1 :],
                    event.target[1][len(self._root_path) + 1 :],
                )

            if event.event == WatchEvent.DIRECTORY_CREATED:
                self._handler.on_directory_created(target)
            elif event.event == WatchEvent.DIRECTORY_REMOVED:
                self._handler.on_directory_removed(target)
            elif event.event == WatchEvent.FILE_CREATED:
                self._handler.on_file_created(target)
            elif event.event == WatchEvent.FILE_MODIFIED:
                self._handler.on_file_modified(target)
            elif event.event == WatchEvent.FILE_REMOVED:
                self._handler.on_file_removed(target)
            elif event.event == WatchEvent.ITEM_MOVED:
                self._handler.on_item_moved(*target)
            else:
                raise NotImplementedError(event.event)
